import asyncio
import tornado
from tornado.websocket import WebSocketHandler
from jupyter_server.utils import url_path_join
from jupyter_server.serverapp import ServerWebApplication
from jupyter_server.base.handlers import APIHandler
from typing import Union, Optional, Awaitable
import base64

from .msgpack_queue import labextension_queue, client_queue
from .tcp_server import tcpServer


settings = None
# {"mode": "proxy", "host": "127.0.0.1", "port": 9001}


class ForwardWebsocketHandler(WebSocketHandler):
    def open(self, *args: str, **kwargs: str) -> Optional[Awaitable[None]]:
        print("Websocket opened for lab extension")
        if settings and settings["mode"] == "direct":
            print("mode is direct, cann't connect jupyter server websocket")
            self.write("mode is direct, cann't connect jupyter server websocket")
            self.close()
            return

        self.task = asyncio.create_task(self.start_loop())

        if not tcpServer.is_running:
            asyncio.create_task(tcpServer.start())

    def on_message(self, message: Union[str, bytes]) -> Optional[Awaitable[None]]:
        buf = message
        buf = base64.standard_b64decode(message)
        # print("put labextension_queue", buf)
        labextension_queue.put(buf)
        # print("write labextension_queue complete")

    def on_close(self) -> None:
        print("Websocket closed for lab extension")
        self.task = None

    async def start_loop(self):
        while self.task:
            while client_queue.qsize() > 0:
                buf = await client_queue.get()
                # print("get client_queue", buf)
                await self.write_message(base64.standard_b64encode(buf))
                # await self.write_message(buf)
            await asyncio.sleep(0.2)


class TcpServerInfoHandler(APIHandler):
    @tornado.web.authenticated
    def get(self):
        if not tcpServer.is_running:
            return self.finish(
                {"code": 1, "message": "tcp server not start, please check server port"}
            )

        addrs = []
        for sock in tcpServer.server.sockets:
            ip, port = sock.getsockname()
            addrs.append(f"{ip}:{port}")

        return self.finish(
            {
                "code": 0,
                "message": "success",
                "data": {"addrs": addrs},
            }
        )


class UpdateSettingsHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        global settings
        settings = tornado.escape.json_decode(self.request.body)
        mode = settings.get("mode", "proxy")

        host = (settings.get("ip", "") or "").strip().split(",")
        while "" in host:
            host.remove("")

        port = settings.get("port", 9001)

        print("-------------debug-------------------")
        settings["mode"] = mode
        settings["host"] = host
        settings["port"] = port
        print(mode, host, port, tcpServer.is_running)
        if mode == "direct":
            if tcpServer.is_running:
                print("tcp server is running, will shutdown")
                asyncio.create_task(tcpServer.stop())
                return self.finish(
                    {
                        "code": 0,
                        "message": "success, tcp server is running, will shutdown",
                    }
                )
            print(" tcp server is closed")
            return self.finish(
                {
                    "code": 0,
                    "message": "success, tcp server is shutdown",
                }
            )

        if host == tcpServer.host and port == tcpServer.port and tcpServer.is_running:
            print("there are not update, ignore")
            return self.finish(
                {
                    "code": 0,
                    "message": "no update, don't restart server",
                }
            )

        tcpServer.host = host
        tcpServer.port = port
        print("will start tcpserver")
        asyncio.create_task(tcpServer.start())
        return self.finish(
            {
                "code": 0,
                "message": "start/restart tcpserver",
            }
        )


def setup_handlers(web_app: ServerWebApplication):
    base_url = web_app.settings["base_url"]
    host_pattern = ".*$"

    def url(sub_route):
        return url_path_join(base_url, "neopyter", sub_route)

    handlers = [
        (url("channel"), ForwardWebsocketHandler),
        (url("get_server_info"), TcpServerInfoHandler),
        (url("update_settings"), UpdateSettingsHandler),
    ]
    web_app.add_handlers(host_pattern, handlers)



from .handler import setup_handlers
from .tcp_server import setup_tcp_server

try:
    from ._version import __version__
except ImportError:
    # Fallback when using the package in dev mode without installing
    # in editable mode with pip. It is highly recommended to install
    # the package from a stable release or in editable mode: https://pip.pypa.io/en/stable/topics/local-project-installs/#editable-installs
    import warnings

    warnings.warn("Importing 'neopyter' outside a proper installation.")
    __version__ = "dev"


__all__ = [
    "_jupyter_server_extension_points",
    "_jupyter_labextension_paths",
    "_load_jupyter_server_extension",
]


def _jupyter_server_extension_points():
    return [{"module": "neopyter"}]


def _jupyter_labextension_paths():
    return [{"src": "labextension", "dest": "neopyter"}]


def _load_jupyter_server_extension(serverapp):
    """
    This function is called when the extension is loaded.
    """
    # print("load jupyter server extension")
    setup_handlers(serverapp.web_app)
    setup_tcp_server(serverapp)



import asyncio
from tornado.websocket import WebSocketClientConnection
from neopyter.msgpack_queue import clear_queue, labextension_queue
import base64
import json


async def test_call_push(jp_fetch, jp_ws_fetch):
    settings = {"mode": "proxy", "host": "127.0.0.1", "port": 9001}
    await clear_queue()
    r = await jp_fetch(
        "neopyter", "update_settings", method="POST", body=json.dumps(settings)
    )
    assert r.code == 200
    res = json.loads(r.body.decode())
    assert res["code"] == 0

    connection: WebSocketClientConnection = await jp_ws_fetch("neopyter", "channel")
    assert isinstance(connection, WebSocketClientConnection)
    await connection.write_message(base64.standard_b64encode("Hello".encode()))
    await asyncio.sleep(1)
    assert labextension_queue.qsize() == 1

    settings["mode"] = "direct"



import socket
from typing import Optional
import asyncio
from asyncio import StreamReader, StreamWriter, Server
from jupyter_server.serverapp import ServerApp
# from jupyterlab_server.app import JupyterLabAp


from .msgpack_queue import clear_queue, labextension_queue, client_queue


async def check_port(host, port):
    try:
        _, w = await asyncio.open_connection(host, port)
        w.close()
        await w.wait_closed()
        return True
    except OSError:
        return False


async def find_empty_port(start_idx):
    idx = start_idx
    while await check_port("127.0.0.1", idx):
        idx = idx + 1
    return idx


def get_all_ip():
    hostname = socket.gethostname()
    ip_addresses = set()
    ip_addresses.add("127.0.0.1")
    for info in socket.getaddrinfo(hostname, None, family=socket.AF_INET):
        ip_addresses.add(info[4][0])
    return ip_addresses


class TcpServer(object):
    def __init__(self) -> None:
        self.server: Optional[Server] = None
        self.builtinHost = get_all_ip()
        self.host = []
        self.port = 9001

    @property
    def is_running(self) -> bool:
        return self.server is not None

    async def start(self):
        # support update?
        if self.server:
            print("stop old server")
            await self.stop()
        host = set.union(self.builtinHost, self.host)
        print("resolved host:", host)
        self.server = await asyncio.start_server(
            lambda r, w: self.client_connected(r, w), list(host), self.port
        )
        addrs = ", ".join(str(sock.getsockname()) for sock in self.server.sockets)
        print(f"Serving on {addrs}")

    async def client_connected(self, reader: StreamReader, writer: StreamWriter):
        if self.server is None:
            writer.close()
            await writer.wait_closed()
            return
        print("New client connected ")
        await clear_queue()
        await asyncio.gather(
            asyncio.create_task(self.start_reader_loop(reader, writer)),
            asyncio.create_task(self.start_writer_loop(writer)),
        )
        print("client disconnected")

    async def start_reader_loop(self, reader: StreamReader, writer: StreamWriter):
        print("Client reader loop start")
        server = self.server
        while not reader.at_eof() and server.sockets:
            buf = await reader.read(512)
            if len(buf) == 0:
                continue
            # print("put client_queue", buf)
            await client_queue.put(buf)
        writer.close()
        await writer.wait_closed()
        print("client reader loop end")

    async def start_writer_loop(self, writer: StreamWriter):
        print("Client writer loop start")
        server = self.server
        while not writer.is_closing() and server.sockets:
            while labextension_queue.qsize() > 0:
                buf = await labextension_queue.get()
                # print("get labextension_queue", buf)
                writer.write(buf)
            await asyncio.sleep(0.3)
        print("client writer loop end")

    async def stop(self):
        # maybe repeat check, whatever
        if self.server:
            self.server.close()
            await self.server.wait_closed()
            self.server = None


# global variable?
tcpServer = TcpServer()


def setup_tcp_server(app: ServerApp):
    global tcpServer



from tornado.queues import Queue
# from asyncio.queues import Queue

# jupyter lab frontend extension => labextension_queue => client(neovim)
labextension_queue = Queue()

# client(neovim) => client_queue => jupyter lab frontend extension
client_queue = Queue()


async def clear_queue():
    while labextension_queue.qsize() > 0:
        await labextension_queue.get()
    while client_queue.qsize() > 0:
        await client_queue.get()
